@@ -63,6 +63,7 @@
 #include "clutter-stage-private.h"
 
 static void clutter_stage_window_iface_init     (ClutterStageWindowIface     *iface);
+static int clutter_stage_gdk_get_scale_factor   (ClutterStageWindow   *stage_window);
 
 static ClutterStageWindowIface *clutter_stage_window_parent_iface = NULL;
 
@@ -122,21 +123,18 @@ clutter_stage_gdk_get_geometry (ClutterStageWindow    *stage_window,
 {
   ClutterStageGdk *stage_gdk = CLUTTER_STAGE_GDK (stage_window);
 
-  if (!stage_gdk->foreign_window)
+  geometry->x = geometry->y = 0;
+
+  if (stage_gdk->window != NULL)
     {
-      if (stage_gdk->window != NULL)
-        {
-          geometry->width = gdk_window_get_width (stage_gdk->window);
-          geometry->height = gdk_window_get_height (stage_gdk->window);
-        }
-      else
-        {
-          geometry->width = 640;
-          geometry->height = 480;
-        }
+      geometry->width = gdk_window_get_width (stage_gdk->window);
+      geometry->height = gdk_window_get_height (stage_gdk->window);
     }
   else
-    clutter_stage_window_parent_iface->get_geometry (stage_window, geometry);
+    {
+      geometry->width = 800;
+      geometry->height = 600;
+    }
 }
 
 static void
@@ -164,10 +162,13 @@ clutter_stage_gdk_resize (ClutterStageWindow *stage_window,
    */
   if (!stage_gdk->foreign_window)
     gdk_window_resize (stage_gdk->window, width, height);
-#if defined(GDK_WINDOWING_WAYLAND)
+#if defined(GDK_WINDOWING_WAYLAND) && defined(COGL_HAS_EGL_PLATFORM_WAYLAND_SUPPORT)
   else if (GDK_IS_WAYLAND_WINDOW (stage_gdk->window))
-    cogl_wayland_onscreen_resize (CLUTTER_STAGE_COGL (stage_gdk)->onscreen,
-                                  width, height, 0, 0);
+    {
+      int scale = clutter_stage_gdk_get_scale_factor (CLUTTER_STAGE_WINDOW (stage_gdk));
+      cogl_wayland_onscreen_resize (CLUTTER_STAGE_COGL (stage_gdk)->onscreen,
+                                    width * scale, height * scale, 0, 0);
+    }
 #endif
 }
 
@@ -207,12 +208,16 @@ clutter_stage_gdk_unrealize (ClutterStageWindow *stage_window)
       stage_gdk->window = NULL;
     }
 
-  clutter_stage_window_parent_iface->unrealize (stage_window);
-
-#if defined(GDK_WINDOWING_WAYLAND)
-  g_clear_pointer (&stage_gdk->subsurface, wl_subsurface_destroy);
-  g_clear_pointer (&stage_gdk->clutter_surface, wl_surface_destroy);
+#if defined(GDK_WINDOWING_WAYLAND) && defined(COGL_HAS_EGL_PLATFORM_WAYLAND_SUPPORT)
+  /* Destroy and clear subsurface as well */
+  if (GDK_IS_WAYLAND_WINDOW (stage_gdk->subsurface))
+    {
+      gdk_window_destroy (stage_gdk->subsurface);
+      stage_gdk->subsurface = NULL;
+    }
 #endif
+
+  clutter_stage_window_parent_iface->unrealize (stage_window);
 }
 
 #if defined(GDK_WINDOWING_WAYLAND)
@@ -220,44 +225,42 @@ static struct wl_surface *
 clutter_stage_gdk_wayland_surface (ClutterStageGdk *stage_gdk)
 {
   GdkDisplay *display;
-  struct wl_compositor *compositor;
-  struct wl_surface *parent_surface;
-  struct wl_region *input_region;
-  gint x, y;
+  cairo_region_t *input_region;
+  cairo_rectangle_int_t empty_rect = { 0, };
+  GdkWindowAttr attributes;
+  gint attributes_mask;
 
   if (!stage_gdk->foreign_window ||
       gdk_window_get_window_type (stage_gdk->window) != GDK_WINDOW_CHILD)
     return gdk_wayland_window_get_wl_surface (stage_gdk->window);
 
-  if (stage_gdk->clutter_surface)
-    return stage_gdk->clutter_surface;
+  if (stage_gdk->subsurface)
+    goto out;
 
   /* On Wayland if we render to a foreign window, we setup our own
    * surface to not render in the same buffers as the embedding
    * framework.
    */
-  display = gdk_display_get_default ();
-  compositor = gdk_wayland_display_get_wl_compositor (display);
-  stage_gdk->clutter_surface = wl_compositor_create_surface (compositor);
+  display = gdk_window_get_display (stage_gdk->window);
+
+  attributes.window_type = GDK_WINDOW_SUBSURFACE;
+  attributes.wclass = GDK_INPUT_OUTPUT;
+  attributes.x = 0;
+  attributes.y = 0;
+  attributes_mask =  GDK_WA_X | GDK_WA_Y;
+
+  stage_gdk->subsurface = gdk_window_new (stage_gdk->window, &attributes, attributes_mask);
+  gdk_window_set_transient_for (stage_gdk->subsurface, stage_gdk->window);
 
   /* Since we run inside GDK, we can let the embedding framework
    * dispatch the events to Clutter. For that to happen we need to
    * disable input on our surface. */
-  input_region = wl_compositor_create_region (compositor);
-  wl_region_add (input_region, 0, 0, 0, 0);
-  wl_surface_set_input_region (stage_gdk->clutter_surface, input_region);
-  wl_region_destroy (input_region);
+  input_region = cairo_region_create_rectangle (&empty_rect);
+  gdk_window_input_shape_combine_region (stage_gdk->subsurface, input_region, 0, 0);
+  cairo_region_destroy (input_region);
 
-  parent_surface = gdk_wayland_window_get_wl_surface (gdk_window_get_toplevel (stage_gdk->window));
-  stage_gdk->subsurface = wl_subcompositor_get_subsurface (stage_gdk->subcompositor,
-                                                           stage_gdk->clutter_surface,
-                                                           parent_surface);
-
-  gdk_window_get_origin (stage_gdk->window, &x, &y);
-  wl_subsurface_set_position (stage_gdk->subsurface, x, y);
-  wl_subsurface_set_desync (stage_gdk->subsurface);
-
-  return stage_gdk->clutter_surface;
+out:
+  return gdk_wayland_window_get_wl_surface (stage_gdk->subsurface);
 }
 #endif
 
@@ -270,30 +273,30 @@ _clutter_stage_gdk_notify_configure (ClutterStageGdk *stage_gdk,
 {
   if (x < 0 || y < 0 || width < 1 || height < 1)
     return;
+
   if (stage_gdk->foreign_window)
     {
-#if defined(GDK_WINDOWING_WAYLAND)
+      ClutterStageCogl *stage_cogl = CLUTTER_STAGE_COGL (stage_gdk);
+      int scale = clutter_stage_gdk_get_scale_factor (CLUTTER_STAGE_WINDOW (stage_gdk));
+
+#if defined(GDK_WINDOWING_WAYLAND) && defined(COGL_HAS_EGL_PLATFORM_WAYLAND_SUPPORT)
       if (GDK_IS_WAYLAND_WINDOW (stage_gdk->window) &&
           gdk_window_get_window_type (stage_gdk->window) == GDK_WINDOW_CHILD &&
           stage_gdk->subsurface)
         {
-          gint rx, ry;
-          gdk_window_get_origin (stage_gdk->window, &rx, &ry);
-          /* TODO: we might need to apply the scale factor here. */
-          wl_subsurface_set_position (stage_gdk->subsurface, rx, ry);
+          cogl_wayland_onscreen_resize (stage_cogl->onscreen,
+                                        width * scale, height * scale, 0, 0);
         }
       else
 #endif
-#if defined(GDK_WINDOWING_X11)
+#if defined(GDK_WINDOWING_X11) && defined(COGL_HAS_XLIB_SUPPORT)
       if (GDK_IS_X11_WINDOW (stage_gdk->window))
         {
-          ClutterStageCogl *stage_cogl = CLUTTER_STAGE_COGL (stage_gdk);
           ClutterBackend *backend = CLUTTER_BACKEND (stage_cogl->backend);
-          int scale_factor = gdk_window_get_scale_factor (stage_gdk->window);
           XConfigureEvent xevent = { ConfigureNotify };
           xevent.window = GDK_WINDOW_XID (stage_gdk->window);
-          xevent.width = width * scale_factor;
-          xevent.height = height * scale_factor;
+          xevent.width = width * scale;
+          xevent.height = height * scale;
 
           /* Ensure cogl knows about the new size immediately, as we will
            * draw before we get the ConfigureNotify response. */
@@ -319,6 +322,7 @@ clutter_stage_gdk_realize (ClutterStageWindow *stage_window)
   gboolean cursor_visible;
   gboolean use_alpha;
   gfloat width, height;
+  int scale;
 
   if (backend->cogl_context == NULL)
     {
@@ -423,9 +427,9 @@ clutter_stage_gdk_realize (ClutterStageWindow *stage_window)
 
   g_object_set_data (G_OBJECT (stage_gdk->window), "clutter-stage-window", stage_gdk);
 
+  scale = clutter_stage_gdk_get_scale_factor (CLUTTER_STAGE_WINDOW (stage_gdk));
   stage_cogl->onscreen = cogl_onscreen_new (backend->cogl_context,
-					    width, height);
-
+                                            width * scale, height * scale);
 #if defined(GDK_WINDOWING_X11) && defined(COGL_HAS_XLIB_SUPPORT)
   if (GDK_IS_X11_WINDOW (stage_gdk->window))
     {
@@ -475,7 +479,9 @@ clutter_stage_gdk_set_fullscreen (ClutterStageWindow *stage_window,
                                   gboolean            is_fullscreen)
 {
   ClutterStageGdk *stage_gdk = CLUTTER_STAGE_GDK (stage_window);
+  ClutterStageCogl *stage_cogl = CLUTTER_STAGE_COGL (stage_window);
   ClutterStage *stage = CLUTTER_STAGE_COGL (stage_window)->wrapper;
+  gboolean swap_throttle;
 
   if (stage == NULL || CLUTTER_ACTOR_IN_DESTRUCTION (stage))
     return;
@@ -489,6 +495,26 @@ clutter_stage_gdk_set_fullscreen (ClutterStageWindow *stage_window,
     gdk_window_fullscreen (stage_gdk->window);
   else
     gdk_window_unfullscreen (stage_gdk->window);
+
+  /* Full-screen stages are usually unredirected to improve performance
+   * by avoiding a copy; when that happens, we need to turn back swap
+   * throttling because we won't be managed by the compositor any more,
+   */
+  swap_throttle = is_fullscreen;
+
+#ifdef GDK_WINDOWING_WAYLAND
+  {
+    /* Except on Wayland, where there's a deadlock due to both Cogl
+     * and GDK attempting to consume the throttling event; see bug
+     * https://bugzilla.gnome.org/show_bug.cgi?id=754671#c1
+     */
+    GdkDisplay *display = clutter_gdk_get_default_display ();
+    if (GDK_IS_WAYLAND_DISPLAY (display))
+      swap_throttle = FALSE;
+  }
+#endif
+
+  cogl_onscreen_set_swap_throttled (stage_cogl->onscreen, swap_throttle);
 }
 
 static void
@@ -570,6 +596,11 @@ clutter_stage_gdk_show (ClutterStageWindow *stage_window,
 
   clutter_actor_map (CLUTTER_ACTOR (CLUTTER_STAGE_COGL (stage_gdk)->wrapper));
 
+#if defined(GDK_WINDOWING_WAYLAND) && defined(COGL_HAS_EGL_PLATFORM_WAYLAND_SUPPORT)
+  if (GDK_IS_WAYLAND_WINDOW (stage_gdk->subsurface))
+    gdk_window_show (stage_gdk->subsurface);
+#endif
+
   /* Foreign window should be shown by the embedding framework. */
   if (!stage_gdk->foreign_window)
     {
@@ -590,6 +621,11 @@ clutter_stage_gdk_hide (ClutterStageWindow *stage_window)
   clutter_actor_unmap (CLUTTER_ACTOR (CLUTTER_STAGE_COGL (stage_gdk)->wrapper));
 
   /* Foreign window should be hidden by the embedding framework. */
+#if defined(GDK_WINDOWING_WAYLAND) && defined(COGL_HAS_EGL_PLATFORM_WAYLAND_SUPPORT)
+  if (GDK_IS_WAYLAND_WINDOW (stage_gdk->subsurface))
+    gdk_window_hide (stage_gdk->subsurface);
+#endif
+
   if (!stage_gdk->foreign_window)
     gdk_window_hide (stage_gdk->window);
 }
@@ -608,6 +644,12 @@ clutter_stage_gdk_get_scale_factor (ClutterStageWindow *stage_window)
   if (stage_gdk->window == NULL)
     return 1;
 
+/* For Wayland using a subsurface, scaling is handled by gdk */
+#if defined(GDK_WINDOWING_WAYLAND) && defined(COGL_HAS_EGL_PLATFORM_WAYLAND_SUPPORT)
+  if (GDK_IS_WAYLAND_WINDOW (stage_gdk->subsurface))
+    return 1;
+  else
+#endif
   return gdk_window_get_scale_factor (stage_gdk->window);
 }
 
@@ -672,6 +714,14 @@ clutter_stage_gdk_dispose (GObject *gobject)
 {
   ClutterStageGdk *stage_gdk = CLUTTER_STAGE_GDK (gobject);
 
+#if defined(GDK_WINDOWING_WAYLAND) && defined(COGL_HAS_EGL_PLATFORM_WAYLAND_SUPPORT)
+  if (GDK_IS_WAYLAND_WINDOW (stage_gdk->subsurface))
+    {
+      gdk_window_destroy (stage_gdk->subsurface);
+      stage_gdk->subsurface = NULL;
+    }
+#endif
+
   if (stage_gdk->window != NULL)
     {
       g_object_set_data (G_OBJECT (stage_gdk->window),
@@ -700,57 +750,9 @@ clutter_stage_gdk_class_init (ClutterStageGdkClass *klass)
   gobject_class->dispose = clutter_stage_gdk_dispose;
 }
 
-#if defined(GDK_WINDOWING_WAYLAND)
-static void
-registry_handle_global (void *data,
-                        struct wl_registry *registry,
-                        uint32_t name,
-                        const char *interface,
-                        uint32_t version)
-{
-  ClutterStageGdk *stage_gdk = data;
-
-  if (strcmp (interface, "wl_subcompositor") == 0)
-    {
-      stage_gdk->subcompositor = wl_registry_bind (registry,
-                                                   name,
-                                                   &wl_subcompositor_interface,
-                                                   1);
-    }
-}
-
-static void
-registry_handle_global_remove (void *data,
-                               struct wl_registry *registry,
-                               uint32_t name)
-{
-}
-
-static const struct wl_registry_listener registry_listener = {
-  registry_handle_global,
-  registry_handle_global_remove
-};
-#endif
-
 static void
 clutter_stage_gdk_init (ClutterStageGdk *stage)
 {
-#if defined(GDK_WINDOWING_WAYLAND)
-  {
-    GdkDisplay *gdk_display = gdk_display_get_default ();
-    if (GDK_IS_WAYLAND_DISPLAY (gdk_display))
-      {
-        struct wl_display *display;
-        struct wl_registry *registry;
-
-        display = gdk_wayland_display_get_wl_display (gdk_display);
-        registry = wl_display_get_registry (display);
-        wl_registry_add_listener (registry, &registry_listener, stage);
-
-        wl_display_roundtrip (display);
-      }
-  }
-#endif
 }
 
 static void
